// @dart=2.9
import 'dart:typed_data';

import 'package:flutter/material.dart';
import "dart:io";

// https://github.com/dart-lang/sdk/issues/45824

//const String HOST = "192.168.1.6";
const String HOST = "192.168.1.80";
const int PORT = 60128;

enum MessageChannelState
{
  IDLE, CONNECTING, RUNNING
}

typedef OnConnected = void Function();

void main() {
  runApp(MyApp());
}

class MyApp extends StatefulWidget
{
  MyApp({Key key}) : super(key: key);

  @override
  MyAppState createState()
  => MyAppState();
}

class MyAppState extends State<MyApp> with WidgetsBindingObserver
{
  MessageChannel _messageChannel;

  @override
  void initState()
  {
    super.initState();
    _connect();
    WidgetsBinding.instance.addObserver(this);
  }

  @override
  void dispose()
  {
    WidgetsBinding.instance.removeObserver(this);
    super.dispose();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state)
  {
    super.didChangeAppLifecycleState(state);
    if (state == AppLifecycleState.resumed)
    {
      _connect();
    }
    else if (_messageChannel != null)
    {
      _messageChannel.stop();
      //while (_messageChannel.isConnected);
      _messageChannel = null;
    }
  }

  void _connect()
  {
    _messageChannel = MessageChannel((){
      setState(() {
        //
      });
    });
    _messageChannel.start(HOST, PORT);
  }

  // This widget is the root of your application.
  @override
  Widget build(BuildContext context) {

    final String conn = (_messageChannel != null && _messageChannel.isConnected ?
      "Connected to " + _messageChannel.getHostAndPort : "Disonnected"
    );

    return MaterialApp(
      title: 'Flutter Demo',
      theme: ThemeData(
        // This is the theme of your application.
        //
        // Try running your application with "flutter run". You'll see the
        // application has a blue toolbar. Then, without quitting the app, try
        // changing the primarySwatch below to Colors.green and then invoke
        // "hot reload" (press "r" in the console where you ran "flutter run",
        // or simply save your changes to "hot reload" in a Flutter IDE).
        // Notice that the counter didn't reset back to zero; the application
        // is not restarted.
        primarySwatch: Colors.blue,
      ),
      home: MyHomePage(title: conn),
    );
  }
}

class MyHomePage extends StatefulWidget {
  MyHomePage({Key key, this.title}) : super(key: key);

  // This widget is the home page of your application. It is stateful, meaning
  // that it has a State object (defined below) that contains fields that affect
  // how it looks.

  // This class is the configuration for the state. It holds the values (in this
  // case the title) provided by the parent (in this case the App widget) and
  // used by the build method of the State. Fields in a Widget subclass are
  // always marked "final".

  final String title;

  @override
  State<MyHomePage> createState() => _MyHomePageState();
}

class _MyHomePageState extends State<MyHomePage> {
  int _counter = 0;

  void _incrementCounter() {
    setState(() {
      // This call to setState tells the Flutter framework that something has
      // changed in this State, which causes it to rerun the build method below
      // so that the display can reflect the updated values. If we changed
      // _counter without calling setState(), then the build method would not be
      // called again, and so nothing would appear to happen.
      _counter++;
    });
  }

  @override
  Widget build(BuildContext context) {
    // This method is rerun every time setState is called, for instance as done
    // by the _incrementCounter method above.
    //
    // The Flutter framework has been optimized to make rerunning build methods
    // fast, so that you can just rebuild anything that needs updating rather
    // than having to individually change instances of widgets.
    return Scaffold(
      appBar: AppBar(
        // Here we take the value from the MyHomePage object that was created by
        // the App.build method, and use it to set our appbar title.
        title: Text(widget.title),
      ),
      body: Center(
        // Center is a layout widget. It takes a single child and positions it
        // in the middle of the parent.
        child: Column(
          // Column is also a layout widget. It takes a list of children and
          // arranges them vertically. By default, it sizes itself to fit its
          // children horizontally, and tries to be as tall as its parent.
          //
          // Invoke "debug painting" (press "p" in the console, choose the
          // "Toggle Debug Paint" action from the Flutter Inspector in Android
          // Studio, or the "Toggle Debug Paint" command in Visual Studio Code)
          // to see the wireframe for each widget.
          //
          // Column has various properties to control how it sizes itself and
          // how it positions its children. Here we use mainAxisAlignment to
          // center the children vertically; the main axis here is the vertical
          // axis because Columns are vertical (the cross axis would be
          // horizontal).
          mainAxisAlignment: MainAxisAlignment.center,
          children: <Widget>[

            const Text(
              'You have pushed the button this many times:',
            ),
            Text(
              '$_counter',
              style: Theme.of(context).textTheme.headline4,
            ),
          ],
        ),
      ),
      floatingActionButton: FloatingActionButton(
        onPressed: _incrementCounter,
        tooltip: 'Increment',
        child: const Icon(Icons.add),
      ), // This trailing comma makes auto-formatting nicer for build methods.
    );
  }
}

class MessageChannel
{
  String _host = "";
  int _port = -1;
  MessageChannelState _state = MessageChannelState.IDLE;
  RawSocket _socket;
  final OnConnected _onConnected;

  MessageChannel(this._onConnected);

  String get getHostAndPort
  => _host + ":" + _port.toString();

  bool get isConnected
  => _state == MessageChannelState.RUNNING;

  void start(String host, int port)
  {
    _host = host;
    _port = port;
    if (_state != MessageChannelState.IDLE)
    {
      return;
    }
    _state = MessageChannelState.CONNECTING;

    RawSocket.connect(_host, _port, timeout: Duration(seconds: 10)).then((RawSocket sock)
    {
      _socket = sock;
      _socket.writeEventsEnabled = false;
      _socket.listen(_onEvent, onDone: () => _onDone(), cancelOnError: false);
      _state = MessageChannelState.RUNNING;
      print("Connected to " + getHostAndPort);
      _onConnected();
    }).catchError((dynamic e)
    {
      _state = MessageChannelState.IDLE;
      print("Disconnected from " + getHostAndPort + ": " + e.toString());
    });
  }

  void stop()
  {
    if (_socket != null)
    {
      _socket.close();
    }
  }

  void _onEvent(RawSocketEvent event)
  {
    switch (event) {
      case RawSocketEvent.read:
        if (_socket.available() > 0)
        {
          print("RawSocketEvent.read: " + _socket.available().toString());
          _onData(_socket.read());
        }
        break;
      case RawSocketEvent.write:
        break;
      case RawSocketEvent.readClosed:
        print("RawSocketEvent.readClosed");
        break;
      case RawSocketEvent.closed:
        break;
      default:
        throw "Unexpected event $event";
    }
  }

  void _onDone()
  {
    _state = MessageChannelState.IDLE;
    _socket = null;
    print("Disconnected from " + getHostAndPort + ": done");
    _onConnected();
  }

  void _onData(Uint8List data)
  {
    String res = "";
    data.forEach((int f)
    {
        res += f.toRadixString(16);
        res += " ";
    });
    print("Data received: [" + res + "]");
  }
}
